from enum import Enum

__all__ = ["LOSS_KEYS"]


class LOSS_KEYS(str, Enum):
    """Module loss keys."""

    RECONSTRUCTION = "reconstruction_loss"
    KL_LOSS = "kl_loss"
    CLASSIFICATION = "classification_loss"
    UNKNOWN_ATTRIBUTE_PENALTY = "unknown_attribute_penalty_loss"



from . import _utils as ul
from ._model import Biolord
from ._module import BiolordClassifyModule, BiolordModule

__all__ = ["Biolord", "BiolordModule", "BiolordClassifyModule"]


def _get_version() -> str:
    import importlib.metadata as importlib_metadata

    return importlib_metadata.version("biolord")


def _setup_logger() -> "logging.Logger":  # noqa: F821
    import logging
    from importlib.metadata import version

    from rich.console import Console
    from rich.logging import RichHandler

    logger = logging.getLogger(__name__)
    # set the logging level
    logger.setLevel(logging.INFO)

    # nice logging outputs
    console = Console(force_terminal=True)
    if console.is_jupyter is True:
        console.is_jupyter = False
    ch = RichHandler(show_path=False, console=console, show_time=False)
    formatter = logging.Formatter("biolord: %(message)s")
    ch.setFormatter(formatter)
    logger.addHandler(ch)

    # this prevents double outputs
    logger.propagate = False
    return logger


__version__ = _get_version()
logger = _setup_logger()

del _get_version, _setup_logger



from typing import Any, Literal, Optional

import numpy as np
import torch
from scvi import REGISTRY_KEYS, settings
from scvi.distributions import NegativeBinomial, Poisson
from scvi.module import Classifier
from scvi.module.base import BaseModuleClass, auto_move_data
from scvi.nn import Decoder, DecoderSCVI, FCLayers
from sklearn.metrics import mean_squared_error, r2_score
from torch import nn
from torch.distributions import Categorical, Normal

from ._constants import LOSS_KEYS

__all__ = ["RegularizedEmbedding", "BiolordModule", "BiolordClassifyModule"]


class RegularizedEmbedding(nn.Module):
    """Regularized embedding module."""

    def __init__(
        self,
        n_input: int,
        n_output: int,
        sigma: float,
        embed: bool = True,
    ):
        super().__init__()
        self.embedding = nn.Embedding(
            num_embeddings=n_input,
            embedding_dim=n_output,
        )
        self.sigma = sigma if embed else 0
        self.embed = embed

    def forward(self, x):
        """Forward pass."""
        x_ = self.embedding(x)
        if self.training and self.sigma != 0:
            noise = torch.zeros_like(x_)
            noise.normal_(mean=0, std=self.sigma)

            x_ = x_ + noise
        x_ = x_ * self.embed
        return x_


class BiolordModule(BaseModuleClass):
    """The :mod:`biolord` module.

    Parameters
    ----------
    n_genes
        Number of input genes.
    n_samples
        Number of layers.
    x_loc
        The expression data location.
    ordered_attributes_map
        Dictionary of ordered classes and their dimensions.
    categorical_attributes_map
        Dictionary for categorical classes, containing categorical values with keys as each category name and values
        as the categorical integer assignment.
    n_latent
        Latent dimension.
    n_latent_attribute_ordered
        Latent dimension of ordered attributes.
    n_latent_attribute_categorical
        Latent dimension of categorical attributes.
    gene_likelihood
        The gene_likelihood model.
    reconstruction_penalty
        MSE error to reconstruction loss.
    use_batch_norm
        Use batch norm in layers.
    use_layer_norm
        Use layer norm in layers.
    unknown_attribute_noise_param
        Noise strength added to encoding of unknown attributes.
    unknown_attributes
        Whether to include learning for unknown attributes
    attribute_dropout_rate
        Dropout rate.
    attribute_nn_width
        Ordered attributes autoencoder layers' width.
    attribute_nn_depth
        Ordered attributes autoencoder number of layers.
    attribute_nn_activation
        Use activation in ordered attributes.
    decoder_width
        Decoder layers' width.
    decoder_depth
        Decoder number of layers.
    decoder_activation
        Use activation in decoder.
    eval_r2_ordered
        Evaluate the R2 w.r.t. the ordered attribute. Set to `True` only if ordered attributes are binned.
    decoder_dropout_rate
        Decoder dropout rate.
    seed
        Random seed.
    """

    def __init__(
        self,
        n_genes: int,
        n_samples: int,
        x_loc: str,
        ordered_attributes_map: Optional[dict[str, int]] = None,
        categorical_attributes_map: Optional[dict[str, dict]] = None,
        n_latent: int = 32,
        n_latent_attribute_categorical: int = 4,
        n_latent_attribute_ordered: int = 16,
        gene_likelihood: Literal["normal", "nb", "poisson"] = "normal",
        reconstruction_penalty: float = 1e2,
        unknown_attribute_penalty: float = 1e1,
        use_batch_norm: bool = True,
        use_layer_norm: bool = False,
        unknown_attribute_noise_param: float = 1e-1,
        unknown_attributes: bool = True,
        attribute_dropout_rate: dict[str, float] = None,
        decoder_width: int = 512,
        decoder_depth: int = 4,
        decoder_activation: bool = True,
        attribute_nn_width: dict[str, int] = None,
        attribute_nn_depth: dict[str, int] = None,
        attribute_nn_activation: bool = True,
        eval_r2_ordered: bool = False,
        decoder_dropout_rate: float = 0.1,
        seed: int = 0,
    ):
        super().__init__()
        gene_likelihood = gene_likelihood.lower()
        assert gene_likelihood in ["normal", "nb", "poisson"], gene_likelihood

        default_width = 256
        default_depth = 2
        torch.manual_seed(seed)
        np.random.seed(seed)
        settings.seed = seed

        self.ae_loss_fn = nn.GaussianNLLLoss()
        self.ae_loss_mse_fn = nn.MSELoss()
        self.reconstruction_penalty = reconstruction_penalty
        self.unknown_attribute_penalty = unknown_attribute_penalty
        self.mm_regression_loss_fn = nn.BCEWithLogitsLoss()

        self.n_genes = n_genes
        self.n_latent = n_latent
        self.x_loc = x_loc
        self.n_latent_attribute_categorical = n_latent_attribute_categorical
        self.n_latent_attribute_ordered = n_latent_attribute_ordered
        self.gene_likelihood = gene_likelihood
        self.use_batch_norm = use_batch_norm
        self.use_layer_norm = use_layer_norm
        self.eval_r2_ordered = eval_r2_ordered

        self.n_decoder_input = n_latent + (
            n_latent_attribute_categorical * len(categorical_attributes_map)
            + n_latent_attribute_ordered * len(ordered_attributes_map)
        )
        self.categorical_attributes_map = (
            categorical_attributes_map if isinstance(categorical_attributes_map, dict) else {}
        )
        self.ordered_attributes_map = ordered_attributes_map if isinstance(ordered_attributes_map, dict) else {}

        if isinstance(attribute_nn_width, dict):
            self.attribute_nn_width = attribute_nn_width
        elif attribute_nn_width is None:
            self.attribute_nn_width = {attribute_: default_width for attribute_ in self.ordered_attributes_map}
        else:
            self.attribute_nn_width = {attribute_: attribute_nn_width for attribute_ in self.ordered_attributes_map}

        if isinstance(attribute_nn_depth, dict):
            self.attribute_nn_depth = attribute_nn_depth
        elif attribute_nn_depth is None:
            self.attribute_nn_depth = {attribute_: default_depth for attribute_ in self.ordered_attributes_map}
        else:
            self.attribute_nn_depth = {attribute_: attribute_nn_depth for attribute_ in self.ordered_attributes_map}

        if isinstance(attribute_dropout_rate, dict):
            self.attribute_dropout_rate = attribute_dropout_rate
        elif attribute_dropout_rate is None:
            self.attribute_dropout_rate = {
                attribute_: decoder_dropout_rate for attribute_ in self.ordered_attributes_map
            }
        else:
            self.attribute_dropout_rate = {
                attribute_: attribute_dropout_rate for attribute_ in self.ordered_attributes_map
            }

        self.latent_codes = RegularizedEmbedding(
            n_input=n_samples, n_output=n_latent, sigma=unknown_attribute_noise_param, embed=unknown_attributes
        )

        # Create Embeddings
        # 1. ordered classes
        reps_ordered = []
        self.ordered_networks = nn.ModuleDict()
        for attribute_, len_ in self.ordered_attributes_map.items():
            if "_rep" in attribute_:
                reps_ordered.append(attribute_)
            else:
                self.ordered_networks[attribute_] = FCLayers(
                    n_in=len_,
                    n_out=self.n_latent_attribute_ordered,
                    n_layers=self.attribute_nn_depth[attribute_],
                    n_hidden=self.attribute_nn_width[attribute_],
                    dropout_rate=self.attribute_dropout_rate[attribute_],
                    bias=False,
                    use_activation=attribute_nn_activation,
                )
        for attribute_ in reps_ordered:
            self.ordered_networks[attribute_] = self.ordered_networks[attribute_.split("_rep")[0]]

        # 2. categorical classes
        self.categorical_embeddings = nn.ModuleDict()
        reps_categorical = []
        for attribute_, unique_categories in self.categorical_attributes_map.items():
            if "_rep" in attribute_:
                reps_categorical.append(attribute_)
            else:
                self.categorical_embeddings[attribute_] = torch.nn.Embedding(
                    len(unique_categories),
                    n_latent_attribute_categorical,
                )
        for attribute_ in reps_categorical:
            self.categorical_embeddings[attribute_] = self.categorical_embeddings[attribute_.split("_rep")[0]]

        # Decoder components
        if self.gene_likelihood in ["nb", "poisson"]:
            self.decoder = DecoderSCVI(
                n_input=self.n_decoder_input,
                n_output=n_genes,
                n_hidden=decoder_width,
                n_layers=decoder_depth,
                use_batch_norm=use_batch_norm,
                use_layer_norm=use_layer_norm,
                scale_activation="softmax",
            )
            self.px_r = torch.nn.Parameter(torch.randn(n_genes))
        else:
            self.decoder = Decoder(
                n_input=self.n_decoder_input,
                n_output=n_genes,
                n_hidden=decoder_width,
                n_layers=decoder_depth,
                use_batch_norm=use_batch_norm,
                use_layer_norm=use_layer_norm,
                use_activation=decoder_activation,
            )

    def _get_inference_input(self, tensors: dict[Any, Any], **kwargs):
        x = tensors[self.x_loc]  # batch_size, n_genes
        sample_indices = tensors[REGISTRY_KEYS.INDICES_KEY].long().ravel()

        categorical_attribute_dict = {}
        for attribute_ in self.categorical_attributes_map:
            categorical_attribute_dict[attribute_] = tensors[attribute_].view(
                -1,
            )

        ordered_attribute_dict = {}
        for attribute_ in self.ordered_attributes_map:
            ordered_attribute_dict[attribute_] = tensors[attribute_]

        input_dict = {
            "genes": x,
            "sample_indices": sample_indices,
            "categorical_attribute_dict": categorical_attribute_dict,
            "ordered_attribute_dict": ordered_attribute_dict,
        }
        return input_dict

    def get_inference_input(self, tensors: dict[Any, Any], **kwargs) -> dict[str, Any]:
        """Convert tensors to valid inference input.

        Parameters
        ----------
        tensors
            Considered inputs.
        kwargs
            Additional arguments

        Returns
        -------
        Dictionary with the module's expected input tensors (`genes`, `sample_indices`, `categorical_attribute_dict`, and `ordered_attribute_dict`).
        """
        return self._get_inference_input(tensors, **kwargs)

    @auto_move_data
    def _inference_attribute_embeddings(
        self,
        genes,
        categorical_attribute_dict,
        ordered_attribute_dict,
        nullify_attribute=None,
    ):
        """Inference over attribute embeddings."""
        nullify_attribute = [] if nullify_attribute is None else nullify_attribute
        inference_output = {}
        batch_size = genes.shape[0]
        for attribute_, embedding_ in self.categorical_embeddings.items():
            latent_i = embedding_(categorical_attribute_dict[attribute_].long())
            latent_i = latent_i.view(batch_size, self.n_latent_attribute_categorical).unsqueeze(
                0
            )  # 1, batch_size, n_latent_attribute_categorical
            if attribute_ in nullify_attribute:
                latent_i = torch.zeros_like(latent_i)
            inference_output[attribute_] = latent_i

        for attribute_, network_ in self.ordered_networks.items():
            latent_i = network_(ordered_attribute_dict[attribute_])
            latent_i = latent_i.view(batch_size, self.n_latent_attribute_ordered).unsqueeze(0)
            if attribute_ in nullify_attribute:
                latent_i = torch.zeros_like(latent_i)
            inference_output[attribute_] = latent_i

        return inference_output

    def _get_latent_unknown_attributes(
        self,
        sample_indices,
    ):
        """Get the module's latent unknown attributes representation."""
        latent_unknown_attributes = self.latent_codes(sample_indices)

        return latent_unknown_attributes

    @auto_move_data
    def inference(
        self,
        genes: torch.Tensor,
        sample_indices: torch.Tensor,
        categorical_attribute_dict: dict[Any, Any],
        ordered_attribute_dict: dict[Any, Any],
        nullify_attribute: Optional[list] = None,
    ) -> dict[str, Any]:
        """Apply module inference.

        Parameters
        ----------
        genes
            Input expression.
        sample_indices
            Indices in the :class:`~anndata.AnnData` object of the input samples.
        categorical_attribute_dict
            Dictionary with categorical attributes as keys and the attribute sample labels as values.
        ordered_attribute_dict
            Dictionary with ordered attributes as keys and the attribute sample values as values.
        nullify_attribute
            Attributes to exclude from inferred latent space.

        Returns
        -------
        Dictionary with the module's expected input tensors (`genes`, `sample_indices`, `categorical_attribute_dict`, and `ordered_attribute_dict`).
        """
        nullify_attribute = [] if nullify_attribute is None else nullify_attribute
        inference_output = {}
        x_ = genes
        library = torch.log(genes.sum(1)).unsqueeze(1)

        latent_unknown_attributes = self._get_latent_unknown_attributes(sample_indices=sample_indices)

        latent_classes = self._inference_attribute_embeddings(
            genes=x_,
            categorical_attribute_dict=categorical_attribute_dict,
            ordered_attribute_dict=ordered_attribute_dict,
            nullify_attribute=nullify_attribute,
        )

        latent_vecs = [latent_unknown_attributes.squeeze()]
        for key_, latent_ in latent_classes.items():
            latent_vecs.append(latent_.squeeze())
            inference_output[key_] = latent_.squeeze()

        latent = torch.cat(latent_vecs, dim=-1)

        inference_output["latent"] = latent
        inference_output["latent_unknown_attributes"] = latent_unknown_attributes
        inference_output["library"] = library

        return inference_output

    def _get_generative_input(self, tensors, inference_outputs, **kwargs):
        input_dict = {
            "latent": inference_outputs["latent"],
            "library": inference_outputs["library"],
        }
        return input_dict

    @auto_move_data
    def generative(
        self,
        latent: torch.Tensor,
        library: torch.Tensor = None,
    ) -> dict[str, Any]:
        """Runs the generative step.

        Parameters
        ----------
        latent
            The concatenated decomposed latent space.
        library
            Library sizes for each cell.

        Returns
        -------
        Dictionary with the generative predictions of the expression distribution.
        """
        if self.gene_likelihood in ["nb", "poisson"]:
            px_scale, _, px_rate, _ = self.decoder(
                dispersion="gene",
                z=latent,
                library=library,
            )
            px_r = torch.exp(self.px_r)
            px = (
                NegativeBinomial(mu=px_rate, theta=px_r, scale=px_scale)
                if self.gene_likelihood == "nb"
                else Poisson(px_rate)
            )  # , scale=px_scale)

            return {
                "means": px.mean,
                "variances": px.variance,
                "distribution": px,
                "samples": px.sample().squeeze(0),
            }

        else:
            p_m, p_v = self.decoder(x=latent)
            px = Normal(loc=p_m, scale=p_v.sqrt())
            return {
                "means": px.loc,
                "variances": px.variance,
                "distribution": px,
                "samples": px.sample().squeeze(0),
            }

    @auto_move_data
    def loss(
        self,
        tensors: dict[str, torch.Tensor],
        inference_outputs: dict[Literal["latent_unknown_attributes"], torch.Tensor],
        generative_outputs: dict[Literal["distribution", "means", "variances"], torch.Tensor],
    ) -> dict[str, float]:
        """Computes the module's loss.

        Parameters
        ----------
        tensors
            Considered model inputs.
        inference_outputs
            Inference step outputs.
        generative_outputs
            Generative step outputs.

        Returns
        -------
        The loss elements.
        """
        x_ = tensors[self.x_loc]
        means = generative_outputs["means"]
        variances = generative_outputs["variances"]

        if self.gene_likelihood in ["nb", "poisson"]:
            reconstruction_loss = -generative_outputs["distribution"].log_prob(x_).sum(-1)
            reconstruction_loss = reconstruction_loss.mean()
        else:
            reconstruction_loss = self.ae_loss_fn(input=means, target=x_, var=variances)

        reconstruction_loss += self.reconstruction_penalty * self.ae_loss_mse_fn(input=means, target=x_)

        unknown_attribute_penalty_loss_val = self.unknown_attribute_penalty_loss(
            inference_outputs["latent_unknown_attributes"]
        )

        return {
            LOSS_KEYS.RECONSTRUCTION: reconstruction_loss,
            LOSS_KEYS.UNKNOWN_ATTRIBUTE_PENALTY: unknown_attribute_penalty_loss_val,
        }

    @staticmethod
    def unknown_attribute_penalty_loss(latent_unknown_attributes: torch.Tensor) -> float:
        """Computes the content penalty term in the loss."""
        return torch.sum(latent_unknown_attributes**2, dim=1).mean()

    @torch.no_grad()
    def r2_metric(
        self,
        tensors: dict[str, torch.Tensor],
        generative_outputs: dict[str, torch.Tensor],
    ) -> tuple[float, float]:
        """Evaluate the :math:`R^2` metric over gene expression.

        Parameters
        ----------
        tensors
            Considered inputs.
        generative_outputs
            Generative model outputs.

        Returns
        -------
        The :math:`R^2` of the mean and standard deviation predictions of the gene expression.
        """
        x = tensors[self.x_loc].detach().cpu().numpy()  # batch_size, n_genes

        batch_size = x.shape[0]
        indices = torch.zeros(batch_size).to(self.device)
        if self.eval_r2_ordered:
            for ordered_attribute_, len_ in self.ordered_attributes_map.items():
                if len_ > 1:
                    attribute_vals = tensors[ordered_attribute_]  # (batch_size, n_class)
                    indices += (
                        attribute_vals
                        * torch.arange(attribute_vals.shape[1])
                        .view(1, -1)
                        .repeat(batch_size, 1)
                        .to(attribute_vals.device)
                    ).sum(dim=1)
                else:
                    indices += tensors[ordered_attribute_].view(
                        -1,
                    )  # (batch_size,)

        for categorical_attribute_ in self.categorical_attributes_map:
            indices += tensors[categorical_attribute_].view(
                -1,
            )  # (batch_size,)

        unique_indices = indices.unique()

        r2_mean = 0.0
        r2_var = 0.0
        k = 0

        pred_x_mean = (
            torch.nan_to_num(generative_outputs["means"], nan=0, neginf=0, posinf=100).detach().cpu().numpy()
        )  # batch_size, n_genes
        pred_x_var = (
            torch.nan_to_num(generative_outputs["variances"], nan=0, neginf=0, posinf=100).detach().cpu().numpy()
        )  # batch_size, n_genes

        for index in unique_indices:
            index_mask = (indices == index).detach().cpu().numpy()
            if index_mask.sum() > 2:
                x_index = x[index_mask]
                means_index = pred_x_mean[index_mask]
                variances_index = pred_x_var[index_mask]

                true_mean_index = np.nanmean(x_index, axis=0)
                pred_mean_index = np.nanmean(means_index, axis=0)

                true_var_index = np.nanvar(x_index, axis=0)
                pred_var_index = (
                    np.nanvar(means_index, axis=0)
                    if self.gene_likelihood in ["nb", "poisson"]
                    else np.nanmean(variances_index, axis=0)
                )

                r2_mean += r2_score(true_mean_index, pred_mean_index)
                r2_var += r2_score(true_var_index, pred_var_index)
                k += 1
            else:
                continue
        if k > 0:
            return r2_mean / k, r2_var / k
        else:
            return r2_mean, r2_var

    @torch.no_grad()
    def get_expression(
        self, tensors: dict[str, torch.Tensor], **inference_kwargs: Any
    ) -> tuple[torch.tensor, torch.tensor]:
        """Computes gene expression means and standard deviation.

        Parameters
        ----------
        tensors
            Considered inputs.
        inference_kwargs
            Additional arguments.

        Returns
        -------
        Prediction of gene expression mean and standard deviation.
        """
        _, generative_outputs = self.forward(
            tensors,
            compute_loss=False,
            inference_kwargs=inference_kwargs,
        )

        mus = torch.nan_to_num(generative_outputs["means"], nan=0, neginf=0, posinf=100)  # batch_size, n_genes
        stds = torch.nan_to_num(generative_outputs["variances"], nan=0, neginf=0, posinf=100)  # batch_size, n_genes
        return mus, stds


class BiolordClassifyModule(BiolordModule):
    """The `biolord-classify` module.

    A :class:`~biolord.BiolordModule` accompanied by regressors for ordered classes and
    classifiers for categorical classes.

    Parameters
    ----------
    categorical_attributes_missing
        Categorical categories representing un-labeled cells.
    classify_all
        Whether to classify all classes or only semi-supervised classes.
    logits
        Classifier output type.
    bias
        Whether to add bias to the regressor.
    classification_penalty
        Classification penalty strength.
    classifier_nn_width
        Classifier's layer width.
    classifier_nn_depth
        Classifier's number of layers.
    classifier_dropout_rate
        Classifier's dropout rate.
    loss_regression
        Loss function for regressors
    kwargs
        Keyword arguments for :class:`~biolord.BiolordModule`.
    """

    def __init__(
        self,
        categorical_attributes_missing: Optional[dict[str, str]] = None,
        classify_all: bool = False,
        logits: bool = False,
        bias: bool = True,
        classification_penalty: float = 1e-1,
        classifier_penalty: float = 1e-4,
        classifier_nn_width: int = 128,
        classifier_nn_depth: int = 2,
        classifier_dropout_rate: float = 1e-1,
        loss_regression: Literal["normal", "mse"] = "normal",
        **kwargs: Any,
    ):
        super().__init__(**kwargs)

        loss_regression = loss_regression.lower()
        assert loss_regression in ["normal", "mse"], loss_regression

        self.classification_penalty = classification_penalty
        self.classifier_penalty = classifier_penalty
        self.classification_loss_fn = nn.CrossEntropyLoss()
        self.regression_loss_fn = nn.MSELoss() if loss_regression == "mse" else nn.GaussianNLLLoss()
        self.loss_regression = loss_regression
        self.mm_regression_loss_fn = nn.BCEWithLogitsLoss()
        self.classify_all = classify_all

        if isinstance(categorical_attributes_missing, dict):
            self.categorical_attributes_missing = categorical_attributes_missing
        elif categorical_attributes_missing is None:
            self.categorical_attributes_missing = {attribute_: None for attribute_ in self.categorical_attributes_map}
        else:
            self.categorical_attributes_missing = {
                attribute_: categorical_attributes_missing for attribute_ in self.categorical_attributes_map
            }

        self.categorical_embeddings = nn.ModuleDict()
        for attribute_, unique_categories in self.categorical_attributes_map.items():
            if self.categorical_attributes_missing[attribute_] is not None:
                self.categorical_embeddings[attribute_] = torch.nn.Embedding(
                    len(unique_categories),
                    self.n_latent_attribute_categorical,
                    padding_idx=self.categorical_attributes_map[attribute_][
                        self.categorical_attributes_missing[attribute_]
                    ],
                )
            else:
                self.categorical_embeddings[attribute_] = torch.nn.Embedding(
                    len(unique_categories),
                    self.n_latent_attribute_categorical,
                )

        # Create classifiers
        self.ordered_regressors = nn.ModuleDict()
        if self.classify_all:
            if self.loss_regression == "mse":
                self.ordered_regressors = nn.ModuleDict(
                    {
                        attribute_: nn.Linear(
                            in_features=self.n_genes,
                            out_features=len_,
                            bias=bias,
                        )
                        for attribute_, len_ in self.ordered_attributes_map.items()
                    }
                )
            else:
                self.ordered_regressors = nn.ModuleDict(
                    {
                        attribute_: Decoder(
                            n_input=self.n_genes,
                            n_output=len_,
                            n_hidden=classifier_nn_width,
                            n_layers=classifier_nn_depth,
                            use_batch_norm=self.use_batch_norm,
                            use_layer_norm=self.use_layer_norm,
                        )
                        for attribute_, len_ in self.ordered_attributes_map.items()
                    }
                )

        self.categorical_classifiers = nn.ModuleDict()
        for attribute_, unique_categories in self.categorical_attributes_map.items():
            if self.categorical_attributes_missing[attribute_] is not None:
                self.categorical_classifiers[attribute_] = Classifier(
                    n_input=self.n_genes,
                    n_labels=len(unique_categories) - 1,
                    n_hidden=classifier_nn_width,
                    n_layers=classifier_nn_depth,
                    dropout_rate=classifier_dropout_rate,
                    logits=logits,
                )
            elif self.classify_all:
                self.categorical_classifiers[attribute_] = Classifier(
                    n_input=self.n_genes,
                    n_labels=len(unique_categories),
                    n_hidden=classifier_nn_width,
                    n_layers=classifier_nn_depth,
                    dropout_rate=classifier_dropout_rate,
                    logits=logits,
                )

    def _get_inference_input(self, tensors: dict, **kwargs):
        x = tensors[self.x_loc]  # batch_size, n_genes
        sample_indices = tensors[REGISTRY_KEYS.INDICES_KEY].long().ravel()

        classification = self.classify(x)

        categorical_attribute_dict = {}
        for attribute_ in self.categorical_attributes_map:
            categorical_attribute_dict[attribute_] = tensors[attribute_].view(
                -1,
            )
            if self.categorical_attributes_missing[attribute_] is not None:
                idx_ = categorical_attribute_dict[attribute_] == self.categorical_attributes_missing[attribute_]
                categorical_attribute_dict[attribute_][idx_] = classification[attribute_][idx_]

        ordered_attribute_dict = {}
        for attribute_ in self.ordered_attributes_map:
            ordered_attribute_dict[attribute_] = tensors[attribute_]

        input_dict = {
            "genes": x,
            "sample_indices": sample_indices,
            "categorical_attribute_dict": categorical_attribute_dict,
            "ordered_attribute_dict": ordered_attribute_dict,
        }
        return input_dict

    @auto_move_data
    def classify(
        self,
        genes: torch.Tensor,
    ) -> dict[str, torch.Tensor]:
        """Run classification.

        Parameters
        ----------
        genes
            Gene expression used for classification.

        Returns
        -------
        Classification output, probability for each ordered attribute and the regression value
        for ordered attributes
        """
        classification = {}
        for attribute_, classifier_ in self.categorical_classifiers.items():
            classification[attribute_] = classifier_(genes)

        for attribute_, regressor_ in self.ordered_regressors.items():
            classification[attribute_] = regressor_(genes)

        return classification

    @auto_move_data
    def _classification_loss(self, tensors: dict[str, torch.Tensor]):
        """Get module classification loss."""
        x = tensors[self.x_loc]  # batch_size, n_genes

        classification_loss = torch.tensor([0.0]).to(self.device)
        classification = self.classify(x)

        for attribute_ in self.categorical_classifiers:
            attribute_vals = tensors[attribute_].view(-1).long()
            if self.categorical_attributes_missing[attribute_] is not None:
                idx_ = (
                    attribute_vals
                    == self.categorical_attributes_map[attribute_][self.categorical_attributes_missing[attribute_]]
                )
                assignment = Categorical(classification[attribute_][idx_, :])
                loss_fn = nn.CrossEntropyLoss(
                    ignore_index=self.categorical_attributes_map[attribute_][
                        self.categorical_attributes_missing[attribute_]
                    ]
                )
                classification_loss += (
                    loss_fn(
                        classification[attribute_],
                        attribute_vals,
                    )
                    + assignment.entropy().mean()
                )
            else:
                classification_loss += self.classification_loss_fn(
                    classification[attribute_],
                    attribute_vals,
                )

        for attribute_, len_ in self.ordered_attributes_map.items():
            if attribute_ in classification:
                if len_ > 1:
                    if self.loss_regression == "mse":
                        classification_loss += self.regression_loss_fn(
                            classification[attribute_], tensors[attribute_].float()
                        ) + self.classification_penalty * self.mm_regression_loss_fn(
                            classification[attribute_], tensors[attribute_].gt(0).float()
                        )
                    else:
                        classification_loss += self.regression_loss_fn(
                            classification[attribute_][0],
                            tensors[attribute_].float(),
                            classification[attribute_][1],
                        )
                else:
                    classification_loss += self.regression_loss_fn(
                        classification[attribute_][0],
                        tensors[attribute_].float(),
                        classification[attribute_][1],
                    )

        return classification_loss

    @auto_move_data
    def loss(
        self,
        tensors: dict[str, torch.Tensor],
        inference_outputs: dict[Literal["latent_unknown_attributes"], torch.Tensor],
        generative_outputs: dict[Literal["distribution", "means", "variances"], torch.Tensor],
    ) -> dict[str, float]:
        """Compute the loss.

        Parameters
        ----------
        tensors
            Considered model inputs.
        inference_outputs
            Inference step outputs.
        generative_outputs
            Generative step outputs.

        Returns
        -------
        The loss elements.
        """
        losses = super().loss(
            tensors=tensors,
            inference_outputs=inference_outputs,
            generative_outputs=generative_outputs,
        )

        losses[LOSS_KEYS.CLASSIFICATION] = self._classification_loss(tensors)

        return losses

    @torch.no_grad()
    def _classification_accuracy(self, tensors):
        """Get module classification accuracy."""
        accuracy_dict = {}
        accuracy_val = 0
        r2_dict = {}
        r2_val = 0
        mse_dict = {}
        mse_val = 0

        x = tensors[self.x_loc]  # .detach().cpu().numpy()  # batch_size, n_genes

        classification = self.classify(x)

        for attribute_ in self.categorical_classifiers:
            attribute_vals_pred = classification[attribute_].argmax(dim=-1).cpu().numpy()
            attribute_vals = tensors[attribute_].view(-1).long().cpu().numpy()
            if self.categorical_attributes_missing[attribute_] is not None:
                idx_ = (
                    attribute_vals
                    != self.categorical_attributes_map[attribute_][self.categorical_attributes_missing[attribute_]]
                )
                if idx_.sum() > 0:
                    accuracy_dict[attribute_] = np.mean(attribute_vals_pred[idx_] == attribute_vals[idx_])
                    accuracy_val += accuracy_dict[attribute_]
            else:
                accuracy_dict[attribute_] = np.mean(attribute_vals_pred == attribute_vals)
                accuracy_val += accuracy_dict[attribute_]

        for attribute_ in self.ordered_regressors:
            attribute_vals_pred = (
                classification[attribute_] if self.regression_loss == "mse" else classification[attribute_][0]
            )
            attribute_vals = tensors[attribute_].cpu().numpy()
            attribute_vals_pred = (
                torch.nan_to_num(attribute_vals_pred, nan=0, neginf=0, posinf=100).detach().cpu().numpy()
            )

            r2_dict[attribute_] = r2_score(attribute_vals, attribute_vals_pred) if attribute_vals.shape[0] > 2 else 0
            mse_dict[attribute_] = mean_squared_error(attribute_vals, attribute_vals_pred)

            r2_val += r2_dict[attribute_]
            mse_val += mse_dict[attribute_]

        accuracy_mean = 0
        r2_mean = 0
        mse_mean = 0
        if len(accuracy_dict):
            accuracy_mean = accuracy_val / len(accuracy_dict)
        if len(r2_dict):
            r2_mean = r2_val / len(r2_dict)
            mse_mean = mse_val / len(mse_dict)

        return accuracy_dict, accuracy_mean, r2_dict, r2_mean, mse_dict, mse_mean



from typing import Optional

from scvi.data import AnnDataManager
from scvi.dataloaders import DataSplitter


class AnnDataSplitter(DataSplitter):
    """Creates data loaders ``train_set``, ``validation_set``, ``test_set`` using given indices."""

    def __init__(
        self,
        adata_manager: AnnDataManager,
        train_indices,
        valid_indices,
        test_indices,
        pin_memory: bool = False,
        **kwargs,
    ):
        super().__init__(adata_manager=adata_manager, pin_memory=pin_memory)
        self.data_loader_kwargs = kwargs
        self.train_idx = train_indices
        self.val_idx = valid_indices
        self.test_idx = test_indices

    def setup(self, stage: Optional[str] = None):
        """Over-ride parent's setup to preserve split idx."""
        return



import io
import itertools
import logging
import os
from collections.abc import Sequence
from typing import Any, Literal, Optional, Union

import numpy as np
import pandas as pd
import rich
import torch
from anndata import AnnData
from lightning.pytorch.callbacks import ModelCheckpoint
from scvi import REGISTRY_KEYS
from scvi.data import AnnDataManager
from scvi.data.fields import (
    CategoricalObsField,
    LayerField,
    NumericalObsField,
    ObsmField,
)
from scvi.dataloaders import DataSplitter
from scvi.model.base import BaseModelClass
from scvi.train import TrainRunner
from scvi.utils import setup_anndata_dsp
from tqdm import tqdm

from ._data import AnnDataSplitter
from ._module import BiolordClassifyModule, BiolordModule
from ._train import biolordClassifyTrainingPlan, biolordTrainingPlan
from ._utils import repeat_n

logger = logging.getLogger(__name__)
logger.propagate = False
logging_dir = "./biolord_log/"

__all__ = ["Biolord"]


class Biolord(BaseModelClass):
    """The biolord model class.

    Parameters
    ----------
    adata
        Annotated data object.
    model_name
        Name of the model.
    module_params
        Hyperparameters for the model's module initialization, e.g, :class:`~biolord.BiolordModule` or
        :class:`~biolord.BiolordClassifyModule`.
    n_latent
        Number of latent dimensions used for the latent embedding.
    train_classifiers
        Whether to activate a :class:`~biolord.BiolordClassifyModule`.
    split_key
        Key in :attr:`anndata.AnnData.obs` used to split the data between train, test and validation.
    train_split
        Value in :attr:`anndata.AnnData.obs` ``['{split_key}']`` marking the train set.
    valid_split
        Value in :attr:`anndata.AnnData.obs` ``['{split_key}']`` marking the validation set.
    test_split
        Value in :attr:`anndata.AnnData.obs` ``['{split_key}']`` marking the test set.

    Examples
    --------
    .. code-block:: python

        import scanpy as sc
        import biolord

        adata = sc.read(...)
        biolord.Biolord.setup_anndata(
            adata, ordered_attributes_keys=["time"], categorical_attributes_keys=["cell_type"]
        )
        model = biolord.Biolord(adata, n_latent=256, split_key="split")
        model.train(max_epochs=200, batch_size=256)
    """

    def __init__(
        self,
        adata: AnnData,
        model_name: Optional[str] = None,
        module_params: dict[str, Any] = None,
        n_latent: int = 128,
        train_classifiers: bool = False,
        split_key: Optional[str] = None,
        train_split: str = "train",
        valid_split: str = "test",
        test_split: str = "ood",
    ):
        super().__init__(adata)

        self.categorical_attributes_map = {}
        self.ordered_attributes_map = {}
        self.retrieval_attribute_dict = {}
        self.categorical_attributes_missing = self.registry_["setup_args"]["categorical_attributes_missing"]
        self.x_loc = None

        self._set_attributes_maps()

        self.n_latent = n_latent
        self.n_genes = adata.n_vars

        self.split_key = split_key
        self.scores = {}

        self._module = None
        self._training_plan = None
        self._data_splitter = None

        train_indices, valid_indices, test_indices = None, None, None
        if split_key is not None:
            train_indices = np.where(adata.obs.loc[:, split_key] == train_split)[0]
            valid_indices = np.where(adata.obs.loc[:, split_key] == valid_split)[0]
            test_indices = np.where(adata.obs.loc[:, split_key] == test_split)[0]

        self.train_indices = train_indices
        self.valid_indices = valid_indices
        self.test_indices = test_indices
        self.n_samples = adata.n_obs
        self.train_classifiers = train_classifiers

        module_params = module_params if isinstance(module_params, dict) else {}

        if self.train_classifiers:
            self.module = BiolordClassifyModule(
                n_genes=self.n_genes,
                n_samples=self.n_samples,
                x_loc=self.x_loc,
                categorical_attributes_map=self.categorical_attributes_map,
                ordered_attributes_map=self.ordered_attributes_map,
                categorical_attributes_missing=self.categorical_attributes_missing,
                n_latent=self.n_latent,
                **module_params,
            ).float()
        else:
            self.module = BiolordModule(
                n_genes=self.n_genes,
                n_samples=self.n_samples,
                x_loc=self.x_loc,
                ordered_attributes_map=self.ordered_attributes_map,
                categorical_attributes_map=self.categorical_attributes_map,
                n_latent=self.n_latent,
                **module_params,
            ).float()

        self._model_summary_string = self.__class__.__name__
        self._model_name = model_name
        self.init_params_ = self._get_init_params(locals())
        self.epoch_history = None

    def _set_attributes_maps(self):
        """Set attributes' maps."""
        for attribute_ in self.registry_["setup_args"]["categorical_attributes_keys"]:
            self.categorical_attributes_map[attribute_] = {
                c: i
                for i, c in enumerate(
                    self.registry_["field_registries"][attribute_]["state_registry"]["categorical_mapping"]
                )
            }
        for attribute_ in self.registry_["setup_args"]["ordered_attributes_keys"]:
            # validata obs
            if attribute_ in self.adata.obs:
                self.ordered_attributes_map[attribute_] = 1
            elif attribute_ in self.adata.obsm:
                self.ordered_attributes_map[attribute_] = self.adata.obsm[attribute_].shape[1]
            else:
                raise KeyError(f"class {attribute_} not found in `adata.obs` or `adata.obsm`.")

        if self.registry_["setup_args"]["retrieval_attribute_key"] is not None:
            self.retrieval_attribute_dict = {
                "retrieval_attribute_key": len(
                    np.unique(self.adata.obs[self.registry_["setup_args"]["retrieval_attribute_key"]])
                )
            }

        self.x_loc = self.registry_["setup_args"]["FIELD"].attr_name

    @property
    def training_plan(self):
        """The model's training plan."""
        return self._training_plan

    @training_plan.setter
    def training_plan(self, plan):
        self._training_plan = plan

    @property
    def data_splitter(self):
        """Data splitter."""
        return self._data_splitter

    @data_splitter.setter
    def data_splitter(self, data_splitter):
        self._data_splitter = data_splitter

    @property
    def module(self) -> BiolordModule:
        """Model's module."""
        return self._module

    @module.setter
    def module(self, module: BiolordModule):
        self._module = module

    @property
    def model_name(self) -> str:
        """Model's name."""
        return self._model_name

    @model_name.setter
    def model_name(self, model_name: str):
        self._model_name = model_name

    @classmethod
    @setup_anndata_dsp.dedent
    def setup_anndata(
        cls,
        adata: AnnData,
        ordered_attributes_keys: Optional[list[str]] = None,
        categorical_attributes_keys: Optional[list[str]] = None,
        categorical_attributes_missing: Optional[dict[str, str]] = None,
        retrieval_attribute_key: Optional[str] = None,
        layer: Optional[str] = None,
        **kwargs: Any,
    ) -> None:
        """Setup function.

        Parameters
        ----------
        adata
            Annotated data object.
        ordered_attributes_keys
            Valid :attr:`anndata.AnnData.obs` or :attr:`anndata.AnnData.obsm` keys for the ordered attributes.
        categorical_attributes_keys
            Valid :attr:`anndata.AnnData.obs` keys for the categorical attributes.
        categorical_attributes_missing
            Categories representing missing labels. Only used if ``train_classifiers=True``.
        retrieval_attribute_key
            Valid :attr:`anndata.AnnData.obs` key for an attribute to evaluate retrieval performance over.
        layer
            Expression layer in :attr:`anndata.AnnData.layers` to use. If :obj:`None`, use :attr:`anndata.AnnData.X`.
        kwargs
            Keyword arguments for :meth:`~scvi.data.AnnDataManager.register_fields`.

        Returns
        -------
        Nothing, just sets up ``adata``.
        """
        if layer is not None:
            if layer not in adata.layers:
                raise KeyError(f"{layer} is not a valid key in `adata.layers`.")
            logger.info(f"Using data from adata.layers[{layer!r}]")
            FIELD = LayerField(
                registry_key="layers",
                layer=layer,
                is_count_data=True,
            )
        else:
            logger.info("Using data from `adata.X`.")
            FIELD = LayerField(registry_key="X", layer=None, is_count_data=False)

        ordered_attributes_keys = ordered_attributes_keys if isinstance(ordered_attributes_keys, list) else []

        categorical_attributes_keys = (
            categorical_attributes_keys if isinstance(categorical_attributes_keys, list) else []
        )

        if categorical_attributes_missing is not None:
            for attribute_, val_ in categorical_attributes_missing.items():
                if val_ is not None:
                    adata.obs[attribute_] = adata.obs[attribute_].astype("category")
                    cats = adata.obs[attribute_].cat.categories
                    idx_ = cats.isin([val_])
                    ordered = list(cats[~idx_]) + [val_]

                    adata.obs[attribute_] = adata.obs[attribute_].cat.reorder_categories(ordered)

        # set retrieval class
        retrieval_attribute_dict = {}
        if retrieval_attribute_key is not None:
            retrieval_attribute_dict = {"retrieval_attribute_key": len(np.unique(adata.obs[retrieval_attribute_key]))}

        # set ordered classes
        ordered_attributes_obs = []
        ordered_attributes_obsm = []
        for attribute_ in ordered_attributes_keys:
            # validata obs
            if attribute_ in adata.obs:
                ordered_attributes_obs.append(attribute_)
            elif attribute_ in adata.obsm:
                ordered_attributes_obsm.append(attribute_)
            else:
                raise KeyError(f"class {attribute_} not found in `adata.obs` or `adata.obsm`.")

        setup_method_args = cls._get_setup_method_args(**locals())
        adata.obs["_indices"] = np.arange(adata.n_obs)
        anndata_fields = (
            [
                FIELD,
                NumericalObsField(REGISTRY_KEYS.INDICES_KEY, "_indices"),
            ]
            + [
                CategoricalObsField(registry_key=attribute_, attr_key=attribute_)
                for attribute_ in categorical_attributes_keys
            ]
            + [
                NumericalObsField(
                    attribute_,
                    attribute_,
                )
                for attribute_ in ordered_attributes_obs
            ]
            + [
                ObsmField(
                    attribute_,
                    attribute_,
                    is_count_data=False,
                    correct_data_format=True,
                )
                for attribute_ in ordered_attributes_obsm
            ]
        )
        if retrieval_attribute_key is not None:
            anndata_fields += [
                CategoricalObsField(
                    registry_key=retrieval_attribute_key,
                    attr_key=retrieval_attribute_key,
                )
            ]

        adata_manager = AnnDataManager(
            fields=anndata_fields,
            setup_method_args=setup_method_args,
        )
        adata_manager.register_fields(adata, **kwargs)
        cls.register_manager(adata_manager)

    @torch.no_grad()
    def get_latent_representation_adata(
        self,
        adata: Optional[AnnData] = None,
        indices: Optional[Sequence[int]] = None,
        batch_size: Optional[int] = 512,
        nullify_attribute: Optional[list[str]] = None,
    ) -> tuple[AnnData, AnnData]:
        """Return the unknown attributes latent space and full latent variable.

        Parameters
        ----------
        adata
            Annotated data object.
        indices
            Optional indices.
        batch_size
            Batch size to use.
        nullify_attribute
            Attribute to nullify in the latent space.

        Returns
        -------
        Two :class:`~anndata.AnnData` objects providing the unknown attributes latent space and
        the concatenated decomposed latent respectively.
        """
        if self.is_trained_ is False:
            raise RuntimeError("Please train the model first.")

        nullify_attribute = [] if nullify_attribute is None else nullify_attribute
        adata = self._validate_anndata(adata)
        if indices is None:
            indices = np.arange(adata.n_obs)
        scdl = self._make_data_loader(adata=adata, indices=indices, batch_size=batch_size, shuffle=False)

        latent_unknown_attributes = []
        latent = []
        for tensors in scdl:
            inference_inputs = self.module.get_inference_input(tensors)
            outputs = self.module.inference(**inference_inputs, nullify_attribute=nullify_attribute)
            latent_unknown_attributes += [outputs["latent_unknown_attributes"].cpu().numpy()]
            latent += [outputs["latent"].cpu().numpy()]

        latent_unknown_attributes_adata = AnnData(
            X=np.concatenate(latent_unknown_attributes, axis=0), obs=adata[indices].obs.copy()
        )
        latent_unknown_attributes_adata.obs_names = adata[indices].obs_names

        latent_adata = AnnData(X=np.concatenate(latent, axis=0), obs=adata[indices].obs.copy())
        latent_adata.obs_names = adata[indices].obs_names

        return latent_unknown_attributes_adata, latent_adata

    @torch.no_grad()
    def get_dataset(
        self,
        adata: Optional[AnnData] = None,
        indices: Optional[Sequence[int]] = None,
    ) -> dict[str, torch.Tensor]:
        """Processes :class:`~anndata.AnnData` object into valid input tensors for the model.

        Parameters
        ----------
        adata
            Annotated data object.
        indices
            Optional indices.

        Returns
        -------
        A dictionary of tensors which can be passed as input to the model.
        """
        adata = self._validate_anndata(adata)
        if indices is None:
            indices = np.arange(adata.n_obs)
        scdl = self._make_data_loader(adata=adata, indices=indices, batch_size=len(indices), shuffle=False)
        return list(scdl)[0]

    @torch.no_grad()
    def predict(
        self,
        adata: Optional[AnnData] = None,
        indices: Optional[Sequence[int]] = None,
        batch_size: Optional[int] = 512,
        nullify_attribute: Optional[list[str]] = None,
    ) -> tuple[AnnData, AnnData]:
        """The model's gene expression prediction for a given :class:`~anndata.AnnData` object.

        Parameters
        ----------
        adata
            Annotated data object.
        indices
            Optional indices.
        batch_size
            Batch size to use.
        nullify_attribute
            Attribute to nullify in latent space.

        Returns
        -------
        Two :class:`~anndata.AnnData` objects representing the model's prediction of the expression mean and variance respectively.
        """
        nullify_attribute = [] if nullify_attribute is None else nullify_attribute
        self.module.eval()

        adata = self._validate_anndata(adata)
        if indices is None:
            indices = np.arange(adata.n_obs)
        scdl = self._make_data_loader(adata=adata, indices=indices, batch_size=batch_size, shuffle=False)
        mus = []
        stds = []
        for tensors in scdl:
            _mus, _stds = self.module.get_expression(tensors, nullify_attribute=nullify_attribute)
            _mus = _mus if _mus.ndim > 1 else _mus[None, :]
            _stds = _stds if _stds.ndim > 1 else _stds[None, :]
            mus.append(_mus.detach().cpu().numpy())
            stds.append(_stds.detach().cpu().numpy())

        pred_adata_mean = AnnData(X=np.concatenate(mus, axis=0), obs=adata.obs.copy())
        pred_adata_var = AnnData(X=np.concatenate(stds, axis=0), obs=adata.obs.copy())

        pred_adata_mean.obs_names = adata.obs_names
        pred_adata_var.obs_names = adata.obs_names

        pred_adata_mean.var_names = adata.var_names
        pred_adata_var.var_names = adata.var_names

        return pred_adata_mean, pred_adata_var

    @torch.no_grad()
    def get_ordered_attribute_embedding(
        self,
        attribute_key: str,
        vals: Optional[Union[float, str, np.ndarray]] = None,
    ) -> np.ndarray:
        """Compute embedding of an ordered attribute.

        Parameters
        ----------
        attribute_key
            The key of the desired attribute.
        vals
            Values of interest.

        Returns
        -------
        Array of the attribute's embedding.
        """
        self.module.eval()
        vals = vals if vals is not None else 1.0
        if isinstance(vals, float):
            batch = torch.tensor([vals]).to(self.device).float()
        elif isinstance(vals, np.ndarray):
            batch = torch.tensor(vals).to(self.device).float()
        else:
            batch = vals

        return self.module.ordered_networks[attribute_key](batch).detach().cpu().numpy()

    @torch.no_grad()
    def get_categorical_attribute_embeddings(
        self, attribute_key: str, attribute_category: Optional[str] = None
    ) -> np.ndarray:
        """Compute embedding of a categorical attribute.

        Parameters
        ----------
        attribute_key
            The key of the desired attribute.
        attribute_category
            A specific category for embedding computation.

        Returns
        -------
        Array of the attribute's embedding.
        """
        if attribute_category is None:
            cat_ids = torch.arange(len(self.categorical_attributes_map[attribute_key]), device=self.device)
        else:
            cat_ids = torch.LongTensor([self.categorical_attributes_map[attribute_key][attribute_category]]).to(
                self.device
            )
        embeddings = self.module.categorical_embeddings[attribute_key](cat_ids).detach().cpu().numpy()

        return embeddings

    def save(
        self,
        dir_path: Optional[str] = None,
        overwrite: bool = False,
        save_anndata: bool = False,
        **anndata_save_kwargs: Any,
    ) -> None:
        """Save the model.

        Parameters
        ----------
        dir_path
            Directory where to save the model. If :obj:`None`, it will be determined automatically.
        overwrite
            Whether to overwrite an existing model.
        save_anndata
            Whether to also save :class:`~anndata.AnnData`.
        anndata_save_kwargs
            Keyword arguments :meth:`scvi.model.base.BaseModelClass.save`.

        Returns
        -------
        Nothing, just saves the model.
        """
        if dir_path is None:
            dir_path = (
                f"./{self.__class__.__name__}_model/"
                if self.model_name is None
                else f"./{self.__class__.__name__}_{self.model_name}_model/"
            )
        super().save(
            dir_path=dir_path,
            overwrite=overwrite,
            save_anndata=save_anndata,
            **anndata_save_kwargs,
        )

        if isinstance(self.training_plan.epoch_history, dict):
            self.epoch_history = pd.DataFrame().from_dict(self.training_plan.epoch_history)
            self.epoch_history.to_csv(os.path.join(dir_path, "history.csv"), index=False)

    @classmethod
    def load(
        cls,
        dir_path: str,
        adata: Optional[AnnData] = None,
        accelerator: str = "auto",
        device: Union[int, list[int], str] = "auto",
        **kwargs: Any,
    ) -> "Biolord":
        """Load a saved model.

        Parameters
        ----------
        dir_path
            Directory where the model is saved.
        adata
            AnnData organized in the same way as data used to train model.
        accelerator
            Supports passing different accelerator types ("cpu", "gpu", "tpu", "ipu", "hpu",
            "mps, "auto") as well as custom accelerator instances.
        device
            The device to use. Can be set to a positive number (int or str), or ``"auto"``
            for automatic selection based on the chosen accelerator.
        kwargs
            Keyword arguments for :meth:`scvi`

        Returns
        -------
        The saved model.
        """
        model = super().load(dir_path, adata, accelerator=accelerator, device=device, **kwargs)

        Biolord.categorical_attributes_map = model.categorical_attributes_map
        Biolord.ordered_attributes_map = model.ordered_attributes_map

        fname = os.path.join(dir_path, "history.csv")
        if os.path.isfile(fname):
            model.epoch_history = pd.read_csv(fname)
        else:
            logger.warning(f"The history file `{fname}` was not found")

        return model

    def train(
        self,
        max_epochs: Optional[int] = None,
        accelerator: str = "auto",
        device: Union[int, list[int], str] = "auto",
        train_size: float = 0.9,
        validation_size: Optional[float] = None,
        plan_kwargs: Optional[dict[str, Any]] = None,
        batch_size: int = 128,
        early_stopping: bool = False,
        **trainer_kwargs: Any,
    ) -> None:
        """Train the :class:`~biolord.Biolord` model.

        Parameters
        ----------
        max_epochs
            Maximum number of epochs for training.
        accelerator
            Supports passing different accelerator types ("cpu", "gpu", "tpu", "ipu", "hpu",
            "mps, "auto") as well as custom accelerator instances.
        device
            The device to use. Can be set to a positive number (int or str), or ``"auto"``
            for automatic selection based on the chosen accelerator.
        train_size
            Fraction of training data in the case of randomly splitting dataset to train/validation
            if :attr:`split_key` is not set in model's constructor.
        validation_size
            Fraction of validation data in the case of randomly splitting dataset to train/validation
            if :attr:`split_key` is not set in model's constructor.
        batch_size
            Size of mini-batches for training.
        early_stopping
            If `True`, early stopping will be used during training on validation dataset.
        plan_kwargs
            Keyword arguments for :class:`~scvi.train.TrainingPlan`.
        trainer_kwargs
            Keyword arguments for :class:`~scvi.train.TrainRunner`.

        Returns
        -------
        Nothing, just trains the :class:`~biolord.Biolord` model.
        """
        plan_kwargs = plan_kwargs if plan_kwargs is not None else {}
        if self.train_classifiers:
            self.training_plan = biolordClassifyTrainingPlan(
                module=self.module,
                **plan_kwargs,
            )
        else:
            self.training_plan = biolordTrainingPlan(
                module=self.module,
                **plan_kwargs,
            )

        monitor = trainer_kwargs.pop("monitor", "val_biolord_metric")
        save_ckpt_every_n_epoch = trainer_kwargs.pop("save_ckpt_every_n_epoch", 20)
        enable_checkpointing = trainer_kwargs.pop("enable_checkpointing", True)

        trainer_kwargs["callbacks"] = [] if "callbacks" not in trainer_kwargs.keys() else trainer_kwargs["callbacks"]
        if enable_checkpointing:
            checkpointing_callback = ModelCheckpoint(monitor=monitor, every_n_epochs=save_ckpt_every_n_epoch)
            trainer_kwargs["callbacks"] += [checkpointing_callback]

        num_workers = trainer_kwargs.pop("num_workers", 0)
        if max_epochs is None:
            n_cells = self.adata.n_obs
            max_epochs = np.min([round((20000 / n_cells) * 400), 400])

        manual_splitting = (
            (self.valid_indices is not None) and (self.train_indices is not None) and (self.test_indices is not None)
        )

        if manual_splitting:
            self.data_splitter = AnnDataSplitter(
                self.adata_manager,
                train_indices=self.train_indices,
                valid_indices=self.valid_indices,
                test_indices=self.test_indices,
                batch_size=batch_size,
                num_workers=num_workers,
            )
        else:
            self.data_splitter = DataSplitter(
                self.adata_manager,
                train_size=train_size,
                validation_size=validation_size,
                batch_size=batch_size,
                num_workers=num_workers,
            )
        es = "early_stopping"
        trainer_kwargs[es] = early_stopping if es not in trainer_kwargs.keys() else trainer_kwargs[es]

        trainer_kwargs["check_val_every_n_epoch"] = trainer_kwargs.get("check_val_every_n_epoch", 1)
        trainer_kwargs["early_stopping_patience"] = trainer_kwargs.get("early_stopping_patience", 20)

        root_dir = logging_dir
        root_dir = (
            os.path.join(root_dir, f"{self.__class__.__name__}/")
            if self.model_name is None
            else os.path.join(root_dir, f"{self.model_name}_{self.__class__.__name__}/")
        )

        trainer_kwargs["default_root_dir"] = trainer_kwargs.pop("default_root_dir", root_dir)

        runner = TrainRunner(
            self,
            training_plan=self.training_plan,
            data_splitter=self.data_splitter,
            max_epochs=max_epochs,
            accelerator=accelerator,
            devices=device,
            early_stopping_monitor=monitor,
            early_stopping_mode="max",
            enable_checkpointing=enable_checkpointing,
            **trainer_kwargs,
        )

        return runner()

    @torch.no_grad()
    def evaluate_retrieval(
        self,
        batch_size: int = None,
        eval_set: Literal["test", "validation"] = "test",
    ) -> float:
        """Returns the accuracy of the retrieval task over the pre-defined `retrieval_class`.

        Parameters
        ----------
        batch_size
            Batch size to use.
        eval_set
            Evaluation dataset.

        Returns
        -------
        Retrieval accuracy over the evaluation dataset.
        """
        k = 1
        if self.is_trained_ is False:
            raise RuntimeError("Please train the model first.")

        batch_size = batch_size if batch_size is not None else self.adata.n_obs

        latent_unknown_attributes_train_adata, _ = self.get_latent_representation_adata(
            adata=self.adata, indices=self.train_indices, batch_size=batch_size
        )

        if eval_set == "test":
            eval_indices = self.test_indices
        elif eval_set == "validation":
            eval_indices = self.valid_indices
        else:
            raise RuntimeError(f"Supports `eval_type` `test` or `validation` but {eval_set} was provided.")

        latent_unknown_attributes_test_adata, _ = self.get_latent_representation_adata(
            adata=self.adata, indices=eval_indices
        )

        return self._retrieval_accuracy(
            latent_unknown_attributes_test_adata.obs[REGISTRY_KEYS.RETRIEVAL_CLASS],
            latent_unknown_attributes_test_adata.X,
            latent_unknown_attributes_train_adata.obs[REGISTRY_KEYS.RETRIEVAL_CLASS],
            latent_unknown_attributes_train_adata.X,
            k=k,
        )

    @torch.no_grad()
    def _retrieval_accuracy(
        self,
        retrieval_class,
        latent_unknown_attributes,
        retrieval_attribute_train,
        latent_unknown_attributes_train,
        k=1,
    ) -> float:
        """Retrieval accuracy score."""
        from sklearn.neighbors import NearestNeighbors

        nbrs = NearestNeighbors(n_neighbors=k).fit(latent_unknown_attributes_train)
        _, ind_ = nbrs.kneighbors(latent_unknown_attributes)

        tot_correct = 0.0
        for i_eval, i_train in enumerate(ind_):
            is_equal = [(retrieval_attribute_train[i_train[i]] == retrieval_class[i_eval]) for i in range(k)]
            is_equal = np.sum(is_equal) / k
            tot_correct = tot_correct + is_equal

        return tot_correct / len(ind_)

    @torch.no_grad()
    def _compute_prediction(
        self,
        adata,
        dataset_source,
        target_attributes,
    ) -> tuple[dict[tuple[Any], Any], Any]:
        """Expression prediction over given inputs.

        Parameters
        ----------
        adata
            An annotated data object containing possible values of the `target_attributes`.
        dataset_source
            Dataset of cells to "shift", make predictions over.
        target_attributes
            Attributes to make predictions over.

        Returns
        -------
        The prediction dict for each attribute value and the original expression prediction.
        """
        keys = list(
            itertools.product(
                *[list(self.categorical_attributes_map[attribute_].keys()) for attribute_ in target_attributes]
            )
        )
        layer = "X" if "X" in dataset_source else "layers"

        pred_original, _ = self.module.get_expression(dataset_source)

        classes_dataset = {}
        predictions_dict = {}

        for attribute_ in target_attributes:
            categories_index = pd.Index(adata.obs[attribute_].values, dtype="category")
            classes_dataset[attribute_] = {}
            for key_, _ in tqdm(zip(*np.unique(categories_index.values, return_counts=True), strict=True)):
                bool_category = categories_index.get_loc(key_)

                adata_cur = adata[bool_category, :].copy()
                dataset = self.get_dataset(adata_cur)

                classes_dataset[attribute_][key_] = dataset[attribute_][0, :]

        for key_ in keys:
            dataset_comb = {}
            n_obs = dataset_source[layer].size(0)
            for key_dataset in dataset_source:
                dataset_comb[key_dataset] = dataset_source[key_dataset].to(self.device)

            for ci, attribute_ in enumerate(target_attributes):
                dataset_comb[attribute_] = repeat_n(classes_dataset[attribute_][key_[ci]], n_obs)

            pred, _ = self.module.get_expression(dataset_comb)
            predictions_dict[key_] = pred

        return predictions_dict, pred_original

    def compute_prediction_adata(
        self,
        adata: AnnData,
        adata_source: AnnData,
        target_attributes: list[str],
        add_attributes: Optional[list[str]] = None,
    ) -> AnnData:
        """Expression prediction over given inputs.

        Parameters
        ----------
        adata
            Annotated data object containing possible values of the ``target_attributes``.
        adata_source
            Annotated data object we wish to make predictions over, e.g., change their ``target_attributes``.
        target_attributes
            Attributes to make predictions over.
        add_attributes
            Additional attributes to add to :attr:`anndata.AnnData.obs` from the original adata
            to the  prediction adata object.

        Returns
        -------
        Annotated data object containing predictions of the cells in all combinations of the ``target_attributes``.
        """
        dataset_source = self.get_dataset(adata_source)

        predictions_dict, _ = self._compute_prediction(
            adata=adata, dataset_source=dataset_source, target_attributes=target_attributes
        )

        preds_ = np.concatenate([val.cpu() for key, val in predictions_dict.items()])
        adata_preds = AnnData(X=preds_, dtype=preds_.dtype)
        for attribute_ in target_attributes:
            adata_preds.obs[attribute_] = "Source"

        start = 0
        obs_names_tmp = adata_preds.obs_names.values
        for key_, vals_ in predictions_dict.items():
            for ci, _ in enumerate(target_attributes):
                adata_preds.obs.iloc[start : start + vals_.shape[0], ci] = key_[ci]
            obs_names_tmp[start : start + vals_.shape[0]] = [
                obs_name + "_" + "_".join([str(k) for k in key_]) for obs_name in adata_source.obs_names
            ]
            start += vals_.shape[0]

        adata_preds.obs_names = obs_names_tmp
        for attribute_ in target_attributes:
            adata_preds.obs[attribute_] = adata_preds.obs[attribute_].astype("category")

        if add_attributes is not None:
            for attribute_ in add_attributes:
                start = 0
                adata_preds.obs[attribute_] = np.nan
                for _ in range(int(adata_preds.shape[0] / adata_source.shape[0])):
                    adata_preds.obs.iloc[start : start + adata_source.shape[0], -1] = adata_source.obs[attribute_]
                    start += adata_source.shape[0]

                adata_preds.obs[attribute_] = adata_preds.obs[attribute_].astype(adata_source.obs[attribute_].dtype)
                if f"{attribute_}_colors" in adata_source.uns:
                    adata_preds.uns[f"{attribute_}_colors"] = adata_source.uns[f"{attribute_}_colors"]

        adata_preds.var_names = adata_source.var_names

        return adata_preds

    def __repr__(self) -> str:
        buffer = io.StringIO()
        summary_string = f"{self._model_summary_string} training status: "
        summary_string += "{}".format("[green]Trained[/]" if self.is_trained else "[red]Not trained[/]")
        console = rich.console.Console(file=buffer)
        with console.capture() as capture:
            console.print(summary_string)
        return capture.get()



from typing import Optional

import anndata
import numpy as np
import pandas as pd
import scanpy as sc
import torch


def repeat_n(x, n, device=None):
    """Returns an n-times repeated version of the Tensor x, repetition dimension is axis 0."""
    device = device if device is not None else "cuda" if torch.cuda.is_available() else "cpu"
    return x.to(device).view(1, -1).repeat(n, 1)


def biolord_metric(
    r2_mean: float,
    r2_var: float,
    classification_accuracy: Optional[float] = np.nan,
) -> float:
    """Evaluate biolord metric.

    Parameters
    ----------
    r2_mean
        r2 score of the mean of the gene expression
    r2_var
        r2 score of the variance of the gene expression
    classification_accuracy
        classification accuracy of classify module

    Returns
    -------
    mean of input values.
    """
    return np.nanmean([r2_mean, r2_var, classification_accuracy])


def compute_uncertainty(
    adata: anndata.AnnData, obs_key: str, use_rep: Optional[str] = None, n_neighbors: Optional[int] = 10
):
    """Evaluate uncertainty of prediction using the given obs.

    Parameters
    ----------
    adata
        Annotated data object with `adata.X` model latent representations.
    obs_key
        Key in :attr:`anndata.AnnData.obs` used to evaluate the uncertainty with respect to.
    use_rep
        Representation to use for neighbors calculation.
    n_neighbors
        Number of neighbors to consider.

    Returns
    -------
    Adds "uncertainty" key to :attr:`anndata.AnnData.obs`.

    Notes
    -----
    Based on chemCPA uncertainty definition.
    """
    sc.pp.neighbors(adata, n_neighbors=n_neighbors, use_rep=use_rep)

    distances = adata.obsp["distances"].A
    neighbours = np.where(distances != 0)

    def entropy(obs, base=None):
        vc = pd.Series(obs).value_counts(normalize=True, sort=False)
        base = np.exp if base is None else base
        return -(vc * np.log(vc) / np.log(base)).sum()

    for i in range(adata.shape[0]):
        cond = neighbours[0] == i
        adata.obs[obs_key].iloc[i]
        key_neigh = adata.obs[obs_key].iloc[neighbours[1][cond]]
        adata.obs.loc[adata.obs_names == adata.obs_names[i], "uncertainty"] = (
            1 / np.log(distances[i].sum()) * entropy(key_neigh, base=2)
        )
    return adata



from scvi.data import synthetic_iid

import biolord


def test_package_has_version():
    biolord.__version__


def test_biolord():
    n_latent = 5
    adata = synthetic_iid()
    biolord.Biolord.setup_anndata(
        adata,
        ordered_attributes_keys=None,
        categorical_attributes_keys=["batch", "labels"],
        retrieval_attribute_key=None,
    )

    model = biolord.Biolord(
        adata=adata,
        n_latent=n_latent,
    )

    model.train(10, check_val_every_n_epoch=1, train_size=0.5, enable_checkpointing=False)

    print(model)



from typing import Optional

import numpy as np
import torch
from scvi.train import TrainingPlan
from torch.optim.lr_scheduler import CosineAnnealingLR, StepLR

from ._constants import LOSS_KEYS
from ._module import BiolordModule
from ._utils import biolord_metric


class biolordTrainingPlan(TrainingPlan):
    """Training plan for the biolord model."""

    def __init__(
        self,
        module: BiolordModule,
        n_steps_kl_warmup: Optional[int] = None,
        n_epochs_kl_warmup: Optional[int] = None,
        n_epochs_warmup: Optional[int] = None,
        checkpoint_freq: int = 20,
        latent_lr=1e-4,
        latent_wd=1e-4,
        decoder_lr=1e-4,
        decoder_wd=1e-4,
        step_size_lr: int = 45,
        batch_size: int = 256,
        cosine_scheduler: bool = False,
        scheduler_max_epochs: int = 1000,
        scheduler_final_lr: float = 1e-5,
        attribute_nn_lr: dict[str, float] = None,
        attribute_nn_wd: dict[str, float] = None,
    ):
        super().__init__(
            module=module,
            lr=latent_lr,
            weight_decay=latent_wd,
            n_steps_kl_warmup=n_steps_kl_warmup,
            n_epochs_kl_warmup=n_epochs_kl_warmup,
            reduce_lr_on_plateau=False,
            lr_factor=None,
            lr_patience=None,
            lr_threshold=None,
            lr_scheduler_metric=None,
            lr_min=None,
        )

        if isinstance(attribute_nn_lr, dict):
            self.attribute_nn_lr = attribute_nn_lr
        elif attribute_nn_lr is None:
            self.attribute_nn_lr = {attribute_: self.latent_lr for attribute_ in self.module.ordered_networks}
        else:
            self.attribute_nn_lr = {attribute_: attribute_nn_lr for attribute_ in self.module.ordered_networks}

        if isinstance(attribute_nn_wd, dict):
            self.attribute_nn_wd = attribute_nn_wd
        elif attribute_nn_wd is None:
            self.attribute_nn_wd = {attribute_: self.latent_wd for attribute_ in self.module.ordered_networks}
        else:
            self.attribute_nn_wd = {attribute_: attribute_nn_wd for attribute_ in self.module.ordered_networks}

        self.n_epochs_warmup = n_epochs_warmup if n_epochs_warmup is not None else 0

        self.decoder_wd = decoder_wd
        self.decoder_lr = decoder_lr

        self.latent_wd = latent_wd
        self.latent_lr = latent_lr

        self.checkpoint_freq = checkpoint_freq

        self.scheduler = CosineAnnealingLR if cosine_scheduler else StepLR
        self.scheduler_params = (
            {"T_max": scheduler_max_epochs, "eta_min": scheduler_final_lr}
            if cosine_scheduler
            else {"step_size": step_size_lr}
        )

        self.step_size_lr = step_size_lr
        self.batch_size = batch_size

        self.automatic_optimization = False
        self.iter_count = 0
        self.training_step_outputs = []
        self.validation_step_outputs = []
        self._epoch_keys = []

        self.epoch_keys = [
            "generative_mean_accuracy",  # accuracy in prediction of mean gene exp.
            "generative_var_accuracy",  # accuracy in prediction of variance of gene exp.
            "biolord_metric",  # combination of metrics above
            LOSS_KEYS.RECONSTRUCTION,
            LOSS_KEYS.UNKNOWN_ATTRIBUTE_PENALTY,
        ]

        self.epoch_history = {"mode": [], "epoch": []}
        for key in self.epoch_keys:
            self.epoch_history[key] = []

    def configure_optimizers(self):
        """Set up optimizers."""
        optimizers = []
        schedulers = []

        # latent codes
        optimizers.append(
            torch.optim.Adam(
                [
                    {
                        "params": list(
                            filter(
                                lambda p: p.requires_grad,
                                self.module.latent_codes.parameters(),
                            )
                        ),
                        "lr": self.latent_lr,
                        "weight_decay": self.latent_wd,
                    }
                ]
            )
        )
        # latent decoder
        optimizers.append(
            torch.optim.Adam(
                [
                    {
                        "params": list(
                            filter(
                                lambda p: p.requires_grad,
                                self.module.decoder.parameters(),
                            )
                        ),
                        "lr": self.decoder_lr,
                        "weight_decay": self.decoder_wd,
                    }
                ]
            )
        )
        # categorical classes
        optimizers.append(
            torch.optim.Adam(
                [
                    {
                        "params": list(
                            filter(
                                lambda p: p.requires_grad,
                                self.module.categorical_embeddings.parameters(),
                            )
                        ),
                        "lr": self.latent_lr,
                        "weight_decay": self.latent_wd,
                    }
                ]
            )
        )
        # ordered classes
        for attribute_, nn_ in self.module.ordered_networks.items():
            params_class = list(filter(lambda p: p.requires_grad, nn_.parameters()))
            optimizers.append(
                torch.optim.Adam(
                    [
                        {
                            "params": params_class,
                            "lr": self.attribute_nn_lr[attribute_],
                            "weight_decay": self.attribute_nn_wd[attribute_],
                        }
                    ]
                )
            )

        if self.step_size_lr is not None:
            for optimizer in optimizers:
                schedulers.append(self.scheduler(optimizer, **self.scheduler_params))
            return optimizers, schedulers
        else:
            return optimizers

    @property
    def epoch_keys(self):
        """Epoch keys getter."""
        return self._epoch_keys

    @epoch_keys.setter
    def epoch_keys(self, epoch_keys: list):
        self._epoch_keys.extend(epoch_keys)

    def training_step(self, batch):
        """Training step."""
        optimizers = self.optimizers()
        # model update
        for optimizer in optimizers:
            optimizer.zero_grad()

        inf_outputs, gen_outputs, losses = self.module.forward(
            batch,
        )

        loss = (
            losses[LOSS_KEYS.RECONSTRUCTION]
            + self.module.unknown_attribute_penalty * losses[LOSS_KEYS.UNKNOWN_ATTRIBUTE_PENALTY]
        )

        self.manual_backward(loss)
        for optimizer in optimizers:
            optimizer.step()

        results = {
            LOSS_KEYS.RECONSTRUCTION: losses[LOSS_KEYS.RECONSTRUCTION].item(),
            LOSS_KEYS.UNKNOWN_ATTRIBUTE_PENALTY: losses[LOSS_KEYS.UNKNOWN_ATTRIBUTE_PENALTY].item(),
        }

        self.iter_count += 1

        for key in self.epoch_keys:
            if key not in results:
                results.update({key: 0.0})

        self.training_step_outputs.append(results)
        return results

    def on_train_epoch_end(self):
        """Training epoch end."""
        outputs = self.training_step_outputs
        self.epoch_history["epoch"].append(self.current_epoch)
        self.epoch_history["mode"].append("train")

        for key in self.epoch_keys:
            self.epoch_history[key].append(np.mean([output[key] for output in outputs]))
            self.log(key, self.epoch_history[key][-1], prog_bar=True)

        if self.current_epoch > 1 and self.current_epoch % self.step_size_lr == 0:
            schedulers = self.lr_schedulers()
            for scheduler in schedulers:
                scheduler.step()

        self.training_step_outputs.clear()

    def validation_step(self, batch, batch_idx):
        """Validation step."""
        inf_outputs, gen_outputs, losses = self(batch)

        r2_mean, r2_var = self.module.r2_metric(batch, gen_outputs)

        results = {}
        for key in losses:
            results.update({key: losses[key].item()})

        results.update({"generative_mean_accuracy": r2_mean})
        results.update({"generative_var_accuracy": r2_var})
        results.update({"biolord_metric": biolord_metric(r2_mean, r2_var)})

        self.validation_step_outputs.append(results)
        return results

    def on_validation_epoch_end(self):
        """Validation step end."""
        outputs = self.validation_step_outputs
        self.epoch_history["epoch"].append(self.current_epoch)
        self.epoch_history["mode"].append("valid")
        for key in self.epoch_keys:
            self.epoch_history[key].append(np.mean([output[key] for output in outputs]))
            self.log(f"val_{key}", self.epoch_history[key][-1], prog_bar=True)

        self.validation_step_outputs.clear()

    def test_step(self, batch, batch_idx):
        """Test step."""
        return self.validation_step(batch, batch_idx)

    def test_epoch_end(self, outputs):
        """Test step end."""
        self.epoch_history["epoch"].append(self.current_epoch)
        self.epoch_history["mode"].append("test")
        for key in self.epoch_keys:
            self.epoch_history[key].append(np.mean([output[key] for output in outputs]))
            self.log(f"test_{key}", self.epoch_history[key][-1], prog_bar=True)


class biolordClassifyTrainingPlan(biolordTrainingPlan):
    """Training plan for the biolord classify model."""

    def __init__(
        self,
        classifier_wd: float = 1e-4,
        classifier_lr: float = 1e-4,
        **kwargs,
    ):
        super().__init__(**kwargs)

        self.classifier_wd = classifier_wd
        self.classifier_lr = classifier_lr

        self.epoch_keys = [
            "classification_accuracy",  # categorical class classification accuracy
            "regression_r2_accuracy",  # ordered class regression r2 accuracy
            "regression_mse",  # ordered class regression mse
            LOSS_KEYS.CLASSIFICATION,
        ]

        self.epoch_history = {"mode": [], "epoch": []}
        for key in self.epoch_keys:
            self.epoch_history[key] = []

    def configure_optimizers(self):
        """Configure optimizers."""
        init_optimizers = 0
        schedulers = []

        if self.step_size_lr is not None:
            optimizers, schedulers = super().configure_optimizers()
            init_optimizers = len(schedulers)
        else:
            optimizers = super().configure_optimizers()
        # add classifiers and regressors
        for _, classifier_ in self.module.categorical_classifiers.items():
            optimizers.append(
                torch.optim.Adam(
                    [
                        {
                            "params": list(
                                filter(
                                    lambda p: p.requires_grad,
                                    classifier_.parameters(),
                                )
                            ),
                            "lr": self.latent_lr,
                            "weight_decay": self.latent_wd,
                            # betas=(0.5, 0.999),
                        }
                    ]
                )
            )

        for _, regressor_ in self.module.ordered_regressors.items():
            optimizers.append(
                torch.optim.Adam(
                    [
                        {
                            "params": list(
                                filter(
                                    lambda p: p.requires_grad,
                                    regressor_.parameters(),
                                )
                            ),
                            "lr": self.latent_lr,
                            "weight_decay": self.latent_wd,
                            # betas=(0.5, 0.999),
                        }
                    ]
                )
            )

        if self.step_size_lr is not None:
            for optimizer in optimizers[init_optimizers:]:
                schedulers.append(self.scheduler(optimizer, **self.scheduler_params))
            return optimizers, schedulers
        else:
            return optimizers

    def training_step(self, batch):
        """Training step."""
        optimizers = self.optimizers()
        # model update
        for optimizer in optimizers:
            optimizer.zero_grad()

        inf_outputs, gen_outputs, losses = self.module.forward(
            batch,
        )

        loss = (
            losses[LOSS_KEYS.RECONSTRUCTION]
            + self.module.unknown_attribute_penalty * losses[LOSS_KEYS.UNKNOWN_ATTRIBUTE_PENALTY]
            + self.module.classifier_penalty * losses[LOSS_KEYS.CLASSIFICATION]
        )

        self.manual_backward(loss)
        for optimizer in optimizers:
            optimizer.step()

        results = {
            LOSS_KEYS.RECONSTRUCTION: losses[LOSS_KEYS.RECONSTRUCTION].item(),
            LOSS_KEYS.UNKNOWN_ATTRIBUTE_PENALTY: losses[LOSS_KEYS.UNKNOWN_ATTRIBUTE_PENALTY].item(),
            LOSS_KEYS.CLASSIFICATION: losses[LOSS_KEYS.CLASSIFICATION].item(),
        }

        self.iter_count += 1

        for key in self.epoch_keys:
            if key not in results:
                results.update({key: 0.0})

        return results

    def validation_step(self, batch, batch_idx):
        """Validation step."""
        results = super().validation_step(batch, batch_idx)

        (
            _,
            classification_accuracy_mean,
            _,
            regression_r2_mean,
            _,
            regression_mse_mean,
        ) = self.module._classification_accuracy(batch)

        results.update({"classification_accuracy": classification_accuracy_mean})
        results.update({"regression_r2_accuracy": regression_r2_mean})
        results.update({"regression_mse": regression_mse_mean})
        results.update(
            {
                "biolord_metric": biolord_metric(
                    results["generative_mean_accuracy"],
                    results["generative_var_accuracy"],
                    classification_accuracy_mean,
                )
            }
        )

        return results
